"""
Reduce
"""

from __future__ import annotations
from typing import Union, TypeVar, Sequence
from returnn.tensor import Tensor, Dim

T = TypeVar("T")

__all__ = [
    "reduce",
    "reduce_sum",
    "reduce_max",
    "reduce_min",
    "reduce_mean",
    "reduce_logsumexp",
    "reduce_any",
    "reduce_all",
    "reduce_argmin",
    "reduce_argmax",
]


def reduce(
    source: Tensor[T],
    *,
    mode: str,
    axis: Union[Dim, Sequence[Dim]],
    use_mask: bool = True,
) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param mode: "sum", "max", "min", "mean", "logsumexp", "any", "all", "argmin", "argmax"
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    # noinspection PyProtectedMember
    return source._raw_backend.reduce(source=source, mode=mode, axis=axis, use_mask=use_mask)


def reduce_sum(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="sum", axis=axis, use_mask=use_mask)


def reduce_max(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="max", axis=axis, use_mask=use_mask)


def reduce_min(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="min", axis=axis, use_mask=use_mask)


def reduce_mean(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="mean", axis=axis, use_mask=use_mask)


def reduce_logsumexp(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="logsumexp", axis=axis, use_mask=use_mask)


def reduce_any(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="any", axis=axis, use_mask=use_mask)


def reduce_all(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="all", axis=axis, use_mask=use_mask)


def reduce_argmin(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="argmin", axis=axis, use_mask=use_mask)


def reduce_argmax(source: Tensor[T], *, axis: Union[Dim, Sequence[Dim]], use_mask: bool = True) -> Tensor[T]:
    """
    Reduce the tensor along the given axis

    :param source:
    :param axis:
    :param use_mask: if True (default), use the time mask (part of dim tag) to ignore padding frames
    :return: tensor with axis removed
    """
    return reduce(source=source, mode="argmax", axis=axis, use_mask=use_mask)



"""
Alternative to the original pprint module.
This one has different behavior for indentation, specifically for dicts.
Also the order of dict items are kept as-is
(which is fine for newer Python versions, which will be the insertion order).
Compare (via our ``pprint``)::
  {
    'melgan': {
      'class': 'subnetwork',
      'from': 'data',
      'subnetwork': {
        'l0': {'class': 'pad', 'mode': 'reflect', 'axes': 'spatial', 'padding': (3, 3), 'from': 'data'},
        'la1': {
          'class': 'conv',
          'from': 'l0',
          'activation': None,
          'with_bias': True,
          'n_out': 384,
          'filter_size': (7,),
          'padding': 'valid',
          'strides': (1,),
          'dilation_rate': (1,)
        },
        'lay2': {'class': 'eval', 'eval': 'tf.nn.leaky_relu(source(0), alpha=0.2)', 'from': 'la1'},
        'layer3_xxx': {
          'class': 'transposed_conv',
          'from': 'lay2',
          'activation': None,
          'with_bias': True,
          'n_out': 192,
          'filter_size': (10,),
          'strides': (5,),
          'padding': 'valid',
          'output_padding': (1,),
          'remove_padding': (3,)
        },
        'output': {'class': 'copy', 'from': 'layer3_xxx'}
      }
    },
    'output': {'class': 'copy', 'from': 'melgan'}
  }
Vs (via original ``pprint``)::
  {'melgan': {'class': 'subnetwork',
              'from': 'data',
              'subnetwork': {'l0': {'axes': 'spatial',
                                    'class': 'pad',
                                    'from': 'data',
                                    'mode': 'reflect',
                                    'padding': (3, 3)},
                             'la1': {'activation': None,
                                     'class': 'conv',
                                     'dilation_rate': (1,),
                                     'filter_size': (7,),
                                     'from': 'l0',
                                     'n_out': 384,
                                     'padding': 'valid',
                                     'strides': (1,),
                                     'with_bias': True},
                             'lay2': {'class': 'eval',
                                      'eval': 'tf.nn.leaky_relu(source(0), '
                                              'alpha=0.2)',
                                      'from': 'la1'},
                             'layer3_xxx': {'activation': None,
                                            'class': 'transposed_conv',
                                            'filter_size': (10,),
                                            'from': 'lay2',
                                            'n_out': 192,
                                            'output_padding': (1,),
                                            'padding': 'valid',
                                            'remove_padding': (3,),
                                            'strides': (5,),
                                            'with_bias': True},
                             'output': {'class': 'copy', 'from': 'layer3_xxx'}}},
   'output': {'class': 'copy', 'from': 'melgan'}}
This is a very simple implementation.
There are other similar alternatives:
* [Rich](https://github.com/willmcgugan/rich)
* [pprint++](https://github.com/wolever/pprintpp)
"""

from __future__ import annotations
from typing import Any
import sys
import numpy


def pprint(obj: Any, *, file=None, prefix="", postfix="", line_prefix="", line_postfix="\n") -> None:
    """
    Pretty-print a Python object.
    """
    if file is None:
        file = sys.stdout
    if "\n" in line_postfix and _type_simplicity_score(obj) <= _type_simplicity_limit:
        prefix = f"{line_prefix}{prefix}"
        line_prefix = ""
        postfix = postfix + line_postfix
        line_postfix = ""

    def _sub_pprint(obj_: Any, prefix_="", postfix_="", inc_indent=True):
        multi_line = "\n" in line_postfix
        if not multi_line and postfix_.endswith(","):
            postfix_ += " "
        pprint(
            obj_,
            file=file,
            prefix=prefix_,
            postfix=postfix_,
            line_prefix=(line_prefix + "  " * inc_indent) if multi_line else "",
            line_postfix=line_postfix,
        )

    def _print(s: str, is_end: bool = False):
        nonlocal prefix  # no need for is_begin, just reset prefix
        file.write(line_prefix)
        file.write(prefix)
        file.write(s)
        if is_end:
            file.write(postfix)
        file.write(line_postfix)
        if "\n" in line_postfix:
            file.flush()
        prefix = ""

    def _print_list():
        for i_, v_ in enumerate(obj):
            _sub_pprint(v_, postfix_="," if i_ < len(obj) - 1 else "")

    if isinstance(obj, list):
        if len(obj) == 0:
            _print("[]", is_end=True)
            return
        _print("[")
        _print_list()
        _print("]", is_end=True)
        return

    if isinstance(obj, tuple):
        if len(obj) == 0:
            _print("()", is_end=True)
            return
        if len(obj) == 1:
            _sub_pprint(obj[0], prefix_=f"{prefix}(", postfix_=f",){postfix}", inc_indent=False)
            return
        _print("(")
        _print_list()
        _print(")", is_end=True)
        return

    if isinstance(obj, set):
        if len(obj) == 0:
            _print("set()", is_end=True)
            return
        _print("{")
        _print_list()
        _print("}", is_end=True)
        return

    if isinstance(obj, dict):
        if len(obj) == 0:
            _print("{}", is_end=True)
            return
        _print("{")
        for i, (k, v) in enumerate(obj.items()):
            _sub_pprint(v, prefix_=f"{k!r}: ", postfix_="," if i < len(obj) - 1 else "")
        _print("}", is_end=True)
        return

    if isinstance(obj, numpy.ndarray):
        _sub_pprint(
            obj.tolist(),
            prefix_=f"{prefix}numpy.array(",
            postfix_=f", dtype=numpy.{obj.dtype}){postfix}",
            inc_indent=False,
        )
        return

    # fallback
    _print(repr(obj), is_end=True)


def pformat(obj: Any) -> str:
    """
    Pretty-format a Python object.
    """
    import io

    s = io.StringIO()
    pprint(obj, file=s)
    return s.getvalue()


_type_simplicity_limit = 120.0  # magic number


def _type_simplicity_score(obj: Any, _offset=0.0) -> float:
    """
    :param Any obj:
    :param float _offset:
    :return: a score, which is a very rough estimate of len(repr(o)), calculated efficiently
    """
    _spacing = 2.0
    if isinstance(obj, bool):
        return 4.0 + _offset
    if isinstance(obj, (int, numpy.integer)):
        if obj == 0:
            return 1.0 + _offset
        return 1.0 + numpy.log10(abs(obj)) + _offset
    if isinstance(obj, str):
        return 2.0 + len(obj) + _offset
    if isinstance(obj, (float, complex, numpy.number)):
        return len(repr(obj)) + _offset
    if isinstance(obj, (tuple, list, set)):
        for x in obj:
            _offset = _type_simplicity_score(x, _offset=_offset + _spacing)
            if _offset > _type_simplicity_limit:
                break
        return _offset
    if isinstance(obj, dict):
        for x in obj.values():  # ignore keys...
            _offset = _type_simplicity_score(x, _offset=_offset + 10.0 + _spacing)  # +10 for key
            if _offset > _type_simplicity_limit:
                break
        return _offset
    if isinstance(obj, numpy.ndarray):
        _offset += 10.0  # prefix/postfix
        if obj.size * 2.0 + _offset > _type_simplicity_limit:  # too big already?
            return obj.size * 2.0 + _offset
        if str(obj.dtype).startswith("int"):
            a = _type_simplicity_score(numpy.max(numpy.abs(obj))) + _spacing
            return obj.size * a + _offset
        a = max([_type_simplicity_score(x) for x in obj.flatten()]) + _spacing
        return obj.size * a + _offset
    # Unknown object. Fallback > _type_simplicity_limit.
    return _type_simplicity_limit + 1.0 + _offset



"""
Some utilities for internal use
"""

from __future__ import annotations
from typing import Union, Iterable, List
from returnn.tensor import Tensor
from returnn.util.basic import RefIdEq
from .. import frontend_layers as rfl


def unique_tensor_list(tensors: Iterable[Tensor]) -> List[Tensor]:
    """
    :param list[Tensor] tensors:
    :return: list with unique tensors
    :rtype: list[Tensor]
    """
    seen = set()  # over ref, not tensor (which is not hashable)
    out = []
    for tensor in tensors:
        if RefIdEq(tensor) not in seen:
            out.append(tensor)
            seen.add(RefIdEq(tensor))
    return out


def copy(tensor: Tensor[rfl.Layer], *, name: Union[rfl.Layer, str]) -> Tensor[rfl.Layer]:
    """copy"""
    return rfl.make_layer({"class": "copy", "from": tensor}, name=name)



"""
Provide :class:`RawWavDataset`.
"""

from __future__ import annotations

import h5py
from .cached2 import CachedDataset2
from returnn.datasets.basic import DatasetSeq
from returnn.log import log
import tempfile
import numpy as np
import time


class RawWavDataset(CachedDataset2):
    """
    This dataset returns the raw waveform information of wav files as sequence input data
    It uses temporary hdf files to buffer the data, to avoid repeatedly reading the
    wav files.
    """

    # Need to keep names as-is for compatibility.
    # noinspection PyPep8Naming
    def __init__(self, listFile, frameLength, frameShift, num_outputs=None, **kwargs):
        """
        constructor

        :type listFile: string
        :param listFile: path to the file containing a list of wav file pathes (on path per line)
                         each line needs to contain exactly one wav file which is considered a sequence
        :type frameLength: int
        :param frameLength: length of one frame in samples
        :type frameShift: int
        :param frameShift: shift length of frame in samples
        :type num_outputs: int
        :param num_outputs: this needs to be set if the data set is used with
                            only input data (e.g. for the extraction
                            process).
        """
        self._flag_buffering = False
        super(RawWavDataset, self).__init__(**kwargs)
        self._listFile = listFile
        with open(self._listFile, "r") as f:
            self._wavFiles = f.readlines()
        self._wavFiles = [l.strip() for l in self._wavFiles]
        self._frameLength = frameLength
        self._frameShift = frameShift
        self._flag_pad = True  # specifies if signal is getting cut or zero padded for last frame

        self._num_seqs = len(self._wavFiles)
        self._seq_index_list = None

        self._hdfBufferHandler, self._hdfBufferPath = self._open_hdf_buffer()

        self.num_inputs = self._frameLength
        self.num_outputs = self._get_num_outputs(num_outputs)

    def _collect_single_seq(self, seq_idx):
        """
        returns the sequence specified by the index seq_idx

        :type seq_idx: int
        :rtype: DatasetSeq | None
        :returns DatasetSeq or None if seq_idx >= num_seqs.
        """
        wav_file_id = self._seq_index_list[seq_idx]
        if not self._isInBuffer(wav_file_id):
            self._load_wav_file_id_into_buffer(wav_file_id)

        return self._collect_single_seq_from_buffer(wav_file_id, seq_idx)

    def _collect_single_seq_from_buffer(self, wav_file_id, seq_idx):
        """
        returns the sequence specified by the index seq_idx

        :type wav_file_id: int
        :type seq_idx: int
        :rtype: DatasetSeq | None
        :returns DatasetSeq or None if seq_idx >= num_seqs.
        """
        inputFeatures = self._get_input_features(wav_file_id)
        outputFeatures = self._get_output_features(wav_file_id)
        inputFeatures = inputFeatures.astype(np.float32)
        if outputFeatures is not None:
            outputFeatures = outputFeatures.astype(np.float32)
        return DatasetSeq(seq_idx, inputFeatures, outputFeatures)

    def _get_num_outputs(self, num_outputs):
        """
        #TBD !!!
        """
        if num_outputs is None:
            raise ValueError("If no output data is used, num_outputs needs to be set for RawWavDataset")
        ret_num_outputs = {"classes": (num_outputs, 2)}
        return ret_num_outputs

    def _get_input_features(self, wavFileId):
        """

        :type wavFileId: int
        :param wavFileId: list index of wav file for which to return the input features
        :rtype: 2D numpy.ndarray (frames, features)
        :return: the 2d array containing the time signal segment for each frame
        """
        if not self._isInBuffer(wavFileId):
            self._load_wav_file_id_into_buffer(wavFileId)

        timeSignal = self._hdfBufferHandler["timeSignal"][str(wavFileId)][...]
        frameLength = self._frameLength
        frameShift = self._frameShift
        nrOfFrames = int(np.ceil((float(timeSignal.shape[0] - frameLength) / frameShift) + 1))
        if self._flag_pad:
            padLength = (nrOfFrames - 1) * frameShift + frameLength - timeSignal.shape[0]
            timeSignalPad = np.zeros((timeSignal.shape[0] + padLength,))
            timeSignalPad[0 : timeSignal.shape[0]] = timeSignal
        else:
            nrOfFrames -= 1
            sigLength = (nrOfFrames - 1) * frameShift + frameLength
            timeSignalPad = np.zeros((sigLength,))
            timeSignalPad[:] = timeSignal[0:sigLength]

        inputFeatures = np.zeros((nrOfFrames, frameLength), dtype=np.float32)
        for i1 in range(nrOfFrames):
            inputFeatures[i1, :] = timeSignalPad[i1 * frameShift : (i1 * frameShift + frameLength)]
        return inputFeatures

    def _get_output_features(self, wav_file_id):
        """

        :type wav_file_id: int
        :param wav_file_id: list index of wav file for which to return the output features
        :rtype: #TBD !!!
        :return: #TBD !!!
        """
        if not self._isInBuffer(wav_file_id):
            self._load_wav_file_id_into_buffer(wav_file_id)
        if not str(wav_file_id) in self._hdfBufferHandler["outputs"].keys():
            return None
        else:
            # TBD !!!
            pass

    def _isInBuffer(self, wavFileId):
        """
        returns true if the wav file has already been loaded into the hdf file buffer

        :type wavFileId: int
        :rtype: bool
        """
        if str(wavFileId) in self._hdfBufferHandler["timeSignal"].keys():
            return True
        else:
            return False

    def _load_wav_file_id_into_buffer(self, wavFileId):
        """
        loads the specified wav file into the hdf file buffer

        :type wavFileId: int
        :param wavFileId: the list index specifying the wav file to be loaded to the buffer
        """
        if self._flag_buffering:
            time.sleep(3)
        self._flag_buffering = True
        if self._isInBuffer(wavFileId):
            return False
        wav_file_path = self._wavFiles[wavFileId]
        import scipy.io.wavfile

        (r, x) = scipy.io.wavfile.read(wav_file_path)
        self._hdfBufferHandler["timeSignal"].create_dataset(str(wavFileId), data=x.astype(np.float32))
        self._flag_buffering = False
        return True

    def _open_hdf_buffer(self):
        """
        opens creates a local hdf file used as buffer to avoid reopening wav files

        :rtype: (h5py._hl.file.File, string)
        :return: (hdf buffer file handler, path to tmp file)
        """
        f_id, tmp_hdf_file_path = tempfile.mkstemp(suffix=".hdf")
        file_handler = h5py.File(tmp_hdf_file_path, "w")
        file_handler.create_group("timeSignal")
        file_handler.create_group("outputs")

        return file_handler, tmp_hdf_file_path

    def get_data_dim(self, key):
        """This is copied from CachedDataset2 but the assertion is
        removed (see CachedDataset2.py)

        :type key: str
        :rtype: int
        :return: number of classes, no matter if sparse or not
        """
        if key == "data":
            return self.num_inputs
        if key in self.num_outputs:
            d = self.num_outputs[key][0]
            return d
        self._load_something()
        if len(self.added_data[0].get_data(key).shape) == 1:
            return super(CachedDataset2, self).get_data_dim(key)  # unknown
        assert len(self.added_data[0].get_data(key).shape) == 2
        return self.added_data[0].get_data(key).shape[1]

    def init_seq_order(self, epoch=None, seq_list=None, seq_order=None):
        """
        :type epoch: int|None
        :param epoch: epoch number
        :param list[str]|None seq_list:
        :param list[int]|None seq_order:
        :param seq_list: only None is currently supported
        Initialize lists:
          self.seq_index  # sorted seq idx
        """
        super(RawWavDataset, self).init_seq_order(epoch=epoch, seq_list=seq_list, seq_order=seq_order)

        if epoch is None:
            self._seq_index_list = range(self.num_seqs)
            return True

        if seq_list is not None or seq_order is not None:
            raise NotImplementedError("init_seq_order of RawWavDataset does not support a predefined seq_list yet.")
        else:
            seq_index = self.get_seq_order_for_epoch(
                epoch, self.num_seqs, lambda s: self.get_seq_length(s).get("data", None)
            )

        self._seq_index_list = seq_index
        if epoch is not None:
            # Give some hint to the user in case he is wondering why the cache is reloading.
            print("Reinitialize dataset seq order for epoch %i." % epoch, file=log.v4)

        return True

    @property
    def num_seqs(self):
        """returns the number of sequences of the dataset

        :rtype: int
        """
        if self._num_seqs is None:
            self._num_seqs = len(self._wavFiles)
        return self._num_seqs



"""
Convert literal Python (Python code which can be parsed with ``ast.literal_eval``
to Pickle.

Code partly taken from here:
https://github.com/albertz/literal-python-to-pickle
"""

import pickle
import ctypes
import os
from .basic import NativeCodeCompiler


def literal_eval(s):
    """
    This can be used as an alternative to ``ast.literal_eval``.
    In contrast to ``ast.literal_eval``, it also accepts bytes,
    and it should be ~5x faster.

    :param str|bytes s:
    :return: any object
    """
    raw_pickle = py_to_pickle(s)
    return pickle.loads(raw_pickle)


def py_to_pickle(s):
    """
    :param str|bytes s:
    :rtype: bytes
    """
    if isinstance(s, bytes):
        in_bytes = s
    else:
        assert isinstance(s, str)
        in_bytes = s.encode("utf8")
    in_ = ctypes.create_string_buffer(in_bytes)
    in_len = len(in_bytes)
    out_len = in_len + 1000  # should always be enough (some buffer len + len of literal Python code)
    out_ = ctypes.create_string_buffer(out_len)

    lib = ctypes.CDLL(_get_native_lib_filename())
    lib.py_to_pickle.argtypes = (ctypes.c_char_p, ctypes.c_size_t, ctypes.c_char_p, ctypes.c_size_t)
    lib.py_to_pickle.restype = ctypes.c_int

    res = lib.py_to_pickle(in_, in_len, out_, out_len)
    assert res == 0, "there was some error"
    return out_.raw


_my_dir = os.path.dirname(os.path.abspath(__file__))
_native_cpp_filename = _my_dir + "/py-to-pickle.cpp"
_native_lib_filename = None


def _get_native_lib_filename():
    """
    :return: path to our patch_atfork lib. see :func:`maybe_restart_returnn_with_atfork_patch`
    :rtype: str
    """
    global _native_lib_filename
    if _native_lib_filename:
        return _native_lib_filename
    native = NativeCodeCompiler(
        base_name="pytopickle",
        code_version=1,
        code=open(_native_cpp_filename).read(),
        is_cpp=True,
        c_macro_defines={"LIB": 1},
    )
    _native_lib_filename = native.get_lib_filename()
    return _native_lib_filename



"""
RETURNN frontend (returnn.frontend) utils
"""

from __future__ import annotations
import contextlib
import numpy
from tensorflow.python.util import nest

from returnn.config import Config, global_config_ctx
from returnn.util.pprint import pprint
import returnn.frontend as rf
from returnn.tensor import Tensor, Dim, TensorDict
import returnn.tf.compat as tf_compat
import returnn.torch.frontend as rft
import returnn.tf.frontend_layers as rfl
from returnn.tf.network import TFNetwork
from returnn.torch.data.tensor_utils import tensor_numpy_to_torch_, tensor_torch_to_numpy_


@contextlib.contextmanager
def tf_scope():
    """tf scope"""
    with tf_compat.v1.Graph().as_default(), tf_compat.v1.Session().as_default() as session:
        yield session


def run_model(extern_data: TensorDict, get_model: rf.GetModelFunc, forward_step: rf.StepFunc) -> TensorDict:
    """run"""
    for v in extern_data.data.values():
        _reset_tensor(v)
    rnd = numpy.random.RandomState(42)
    for v in extern_data.data.values():
        _fill_random(v, rnd=rnd)

    with rft.TorchBackend.random_journal_record() as random_journal:
        out_pt = run_model_torch(extern_data, get_model, forward_step)
        # get the values now because dims might get overwritten
        out_pt_raw = out_pt.as_raw_tensor_dict(include_const_sizes=True)

    with rfl.ReturnnLayersBackend.random_journal_replay(random_journal):
        out_tf = run_model_net_dict_tf(extern_data, get_model, forward_step)
        out_tf_raw = out_tf.as_raw_tensor_dict(include_const_sizes=True)

    print(out_pt, out_tf)
    assert set(out_pt.data.keys()) == set(out_tf.data.keys())
    for k, v_pt in out_pt.data.items():
        v_tf = out_tf[k]
        # We cannot really check the dims directly for equality,
        # because the model code often creates new dims, which are different in each call.
        # However, via mark_as_output, the order of dims is well-defined.
        # So we can check the values.
        assert len(v_pt.dims) == len(v_tf.dims)
        assert v_pt.feature_dim_axis == v_tf.feature_dim_axis
        for d_pt, d_tf in zip(v_pt.dims, v_tf.dims):
            assert isinstance(d_pt, Dim) and isinstance(d_tf, Dim)
            assert _dim_is_scalar_size(d_pt) == _dim_is_scalar_size(d_tf)
            if _dim_is_scalar_size(d_pt):
                assert _dim_scalar_size(d_pt) == _dim_scalar_size(d_tf)
            else:
                assert d_pt.dyn_size_ext and d_tf.dyn_size_ext
                # There might be cases where the dims are maybe not equal
                # (same reasoning as above, or also different order),
                # although this would be quite exotic.
                # Let's just assume for now that this does not happen.
                assert d_pt.dyn_size_ext.dims == d_tf.dyn_size_ext.dims
                assert (d_pt.dyn_size_ext.raw_tensor == d_tf.dyn_size_ext.raw_tensor).all()
    assert set(out_pt_raw.keys()) == set(out_tf_raw.keys())
    for k, v_pt in out_pt_raw.items():
        v_tf = out_tf_raw[k]
        assert numpy.allclose(v_pt, v_tf, atol=1e-5, rtol=1e-5)
    return out_pt


def run_model_torch(extern_data: TensorDict, get_model: rf.GetModelFunc, forward_step: rf.StepFunc) -> TensorDict:
    """run"""
    extern_data_raw = extern_data.as_raw_tensor_dict()
    rf.select_backend_torch()
    rf.set_random_seed(42)

    # Inplace replace Numpy by Torch.
    # Inplace because dim tags cannot easily be copied (then they are not the same).
    # We recover extern_data in the end.
    for v in extern_data.data.values():
        tensor_numpy_to_torch_(v)

    model = get_model(epoch=1, step=0)
    rf.init_forward_step_run_ctx()
    forward_step(model=model, extern_data=extern_data)
    outputs = rf.get_run_ctx().outputs
    assert outputs.data
    for v in outputs.data.values():
        tensor_torch_to_numpy_(v)

    extern_data.assign_from_raw_tensor_dict_(extern_data_raw)
    return outputs


def run_model_net_dict_tf(extern_data: TensorDict, get_model: rf.GetModelFunc, forward_step: rf.StepFunc) -> TensorDict:
    """run"""
    extern_data_raw = extern_data.as_raw_tensor_dict()
    for v in extern_data.data.values():
        _reset_tensor(v)
    rf.select_backend_returnn_layers_tf()
    rf.set_random_seed(42)

    from returnn.tf.frontend_layers.config_entry_points import get_net_dict

    config = Config(
        {
            "debug_runtime_sanity_checks": True,
            "extern_data": extern_data,
            "get_model": get_model,
            "task": "forward",
            "forward_step": forward_step,
        }
    )

    with tf_scope() as session, global_config_ctx(config):
        net_dict = get_net_dict(epoch=1, step=0)
        print("*** TF net dict:")
        pprint(net_dict)
        outputs_layers = rf.get_run_ctx().outputs
        print("*** outputs:", outputs_layers)

        net = TFNetwork(config=config, train_flag=False)
        net.construct_from_dict(net_dict)
        session.run(tf_compat.v1.global_variables_initializer())

        outputs_tf = TensorDict()
        for k, v in outputs_layers.data.items():
            v: Tensor[rfl.Layer]
            assert isinstance(v.raw_tensor, rfl.Layer)
            layer_name = v.raw_tensor.get_abs_name()
            layer = net.get_layer(layer_name)
            outputs_tf.data[k] = layer.output.copy()

        fetches = outputs_tf.as_raw_tensor_dict()
        assert set(extern_data.data.keys()) == set(net.extern_data.data.keys())
        extern_data_tf_placeholders = net.extern_data.as_raw_tensor_dict()
        assert set(extern_data_tf_placeholders.keys()) == set(extern_data_raw.keys())
        feed_dict = {extern_data_tf_placeholders[k]: v for k, v in extern_data_raw.items()}

        outputs_numpy_raw = session.run(fetches, feed_dict=feed_dict)

        # Scalars are not Numpy arrays, but our code below assumes that we only have Numpy arrays.
        # So we convert them here.
        def _make_numpy_array(x):
            if isinstance(x, numpy.ndarray):
                return x
            return numpy.array(x)

        outputs_numpy_raw = nest.map_structure(_make_numpy_array, outputs_numpy_raw)

        outputs_numpy = outputs_tf.copy_template()
        for v in outputs_numpy.data.values():
            _reset_tensor(v)
        outputs_numpy.assign_from_raw_tensor_dict_(outputs_numpy_raw)

    extern_data.assign_from_raw_tensor_dict_(extern_data_raw)
    return outputs_numpy


def _reset_tensor(x: Tensor):
    """reset"""
    x.raw_tensor = None
    for dim in x.dims:
        if dim.dyn_size_ext:
            _reset_tensor(dim.dyn_size_ext)


def _fill_random(x: Tensor, *, min_val: int = 0, rnd: numpy.random.RandomState) -> bool:
    """fill. return whether sth was filled"""
    filled = False
    while True:
        have_unfilled = False
        filled_this_round = False

        for dim in x.dims:
            if dim.is_batch_dim() and not dim.dyn_size_ext:
                dim.dyn_size_ext = Tensor("batch", [], dtype="int32")
            if not dim.dyn_size_ext:
                continue
            if _fill_random(dim.dyn_size_ext, min_val=2, rnd=rnd):
                filled = True
                filled_this_round = True
            if dim.dyn_size_ext.raw_tensor is None:
                have_unfilled = True
            elif not isinstance(dim.dyn_size_ext.raw_tensor, numpy.ndarray):
                have_unfilled = True

        if have_unfilled:
            assert filled_this_round, f"should have filled something, {x}"

        if not have_unfilled:
            break

    if x.raw_tensor is not None:
        if not isinstance(x.raw_tensor, numpy.ndarray):
            x.raw_tensor = None

    if x.raw_tensor is None:
        shape = [d.get_dim_value() for d in x.dims]
        if x.dtype.startswith("int"):
            max_val = 10
            if x.sparse_dim and x.sparse_dim.dimension is not None:
                max_val = x.sparse_dim.dimension
            x.raw_tensor = rnd.randint(min_val, max_val, size=shape, dtype=x.dtype)
        elif x.dtype.startswith("float"):
            x.raw_tensor = rnd.normal(0.0, 1.0, size=shape).astype(x.dtype)
        else:
            raise NotImplementedError(f"not implemented for {x} dtype {x.dtype}")
        filled = True

    assert isinstance(x.raw_tensor, numpy.ndarray)

    return filled


def _dim_is_scalar_size(dim: Dim) -> bool:
    """dim is scalar size"""
    if dim.size is not None:
        return True
    if dim.dyn_size_ext:
        return dim.dyn_size_ext.dims == ()
    return False


def _dim_scalar_size(dim: Dim) -> int:
    """dim scalar size"""
    if dim.size is not None:
        return dim.size
    if dim.dyn_size_ext:
        assert dim.dyn_size_ext.dims == ()
        return dim.dyn_size_ext.raw_tensor
    raise Exception(f"dim {dim} has no known size")
